@@ -1,12 +1,6 @@
 #define IN_LIBEXSLT
 #include "libexslt/libexslt.h"
 
-#if defined(_WIN32) && !defined (__CYGWIN__) && (!__MINGW32__)
-#include <win32config.h>
-#else
-#include "config.h"
-#endif
-
 #include <string.h>
 
 #include <libxml/tree.h>
@@ -120,6 +114,7 @@ exsltFuncRegisterImportFunc (void *payload, void *vctxt,
 	    xsltGenericError(xsltGenericErrorContext,
 		    "Failed to register function {%s}%s\n",
 		    URI, name);
+            xmlFree(func);
 	} else {		/* Do the registration */
 	    xsltGenericDebug(xsltGenericDebugContext,
 	            "exsltFuncRegisterImportFunc: register {%s}%s\n",
@@ -191,9 +186,11 @@ exsltFuncShutdown (xsltTransformContextPtr ctxt ATTRIBUTE_UNUSED,
 		   void *vdata) {
     exsltFuncData *data = (exsltFuncData *) vdata;
 
-    if (data->result != NULL)
-	xmlXPathFreeObject(data->result);
-    xmlFree(data);
+    if (data != NULL) {
+        if (data->result != NULL)
+            xmlXPathFreeObject(data->result);
+        xmlFree(data);
+    }
 }
 
 /**
@@ -291,17 +288,13 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
     xmlXPathObjectPtr oldResult, ret;
     exsltFuncData *data;
     exsltFuncFunctionData *func;
-    xmlNodePtr paramNode, oldInsert, fake;
-    int oldBase;
+    xmlNodePtr paramNode, oldInsert, oldXPNode, fake;
+    int oldBase, newBase;
     void *oldCtxtVar;
     xsltStackElemPtr params = NULL, param;
     xsltTransformContextPtr tctxt = xsltXPathGetTransformContext(ctxt);
-    int i, notSet;
-    struct objChain {
-	struct objChain *next;
-	xmlXPathObjectPtr obj;
-    };
-    struct objChain	*savedObjChain = NULL, *savedObj;
+    int i;
+    xmlXPathObjectPtr *args = NULL;
 
     /*
      * retrieve func:function template
@@ -360,6 +353,13 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
     }
     tctxt->depth++;
 
+    /* Evaluating templates can change the XPath context node. */
+    oldXPNode = tctxt->xpathCtxt->node;
+
+    fake = xmlNewDocNode(tctxt->output, NULL,
+			 (const xmlChar *)"fake", NULL);
+    if (fake == NULL)
+        goto error;
     /*
      * We have a problem with the evaluation of function parameters.
      * The original library code did not evaluate XPath expressions until
@@ -382,16 +382,15 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
      * In order to give the function params and variables a new 'scope'
      * we change varsBase in the context.
      */
-    oldBase = tctxt->varsBase;
-    tctxt->varsBase = tctxt->varsNr;
+    newBase = tctxt->varsNr;
     /* If there are any parameters */
     if (paramNode != NULL) {
+        args = (xmlXPathObjectPtr *) xmlMalloc(sizeof(*args) * nargs);
+        if (args == NULL)
+            goto error;
         /* Fetch the stored argument values from the caller */
-	for (i = 0; i < nargs; i++) {
-	    savedObj = xmlMalloc(sizeof(struct objChain));
-	    savedObj->next = savedObjChain;
-	    savedObj->obj = valuePop(ctxt);
-	    savedObjChain = savedObj;
+	for (i = nargs - 1; i >= 0; i--) {
+            args[i] = valuePop(ctxt);
 	}
 
 	/*
@@ -408,17 +407,20 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
 	 * as arguments from the caller
 	 * Calculate the number of un-set parameters
 	 */
-	notSet = func->nargs - nargs;
-	for (; i > 0; i--) {
+	for (i = 0; i < func->nargs; i++) {
 	    param = xsltParseStylesheetCallerParam (tctxt, paramNode);
-	    if (i > notSet) {	/* if parameter value set */
+            if (param == NULL) {
+                xsltLocalVariablePop(tctxt, newBase, -2);
+	        xsltFreeStackElemList(params);
+                for (; i < nargs; i++)
+                    xmlXPathFreeObject(args[i]);
+                goto error;
+            }
+	    if (i < nargs) {	/* if parameter value set */
 		param->computed = 1;
 		if (param->value != NULL)
 		    xmlXPathFreeObject(param->value);
-		savedObj = savedObjChain;	/* get next val from chain */
-		param->value = savedObj->obj;
-		savedObjChain = savedObjChain->next;
-		xmlFree(savedObj);
+		param->value = args[i];
 	    }
 	    xsltLocalVariablePush(tctxt, param, -1);
 	    param->next = params;
@@ -430,11 +432,11 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
      * Actual processing. The context variable is cleared and restored
      * when func:result is evaluated.
      */
-    fake = xmlNewDocNode(tctxt->output, NULL,
-			 (const xmlChar *)"fake", NULL);
+    oldBase = tctxt->varsBase;
     oldInsert = tctxt->insert;
     oldCtxtVar = data->ctxtVar;
     data->ctxtVar = tctxt->contextVariable;
+    tctxt->varsBase = newBase;
     tctxt->insert = fake;
     tctxt->contextVariable = NULL;
     xsltApplyOneTemplate (tctxt, tctxt->node,
@@ -446,6 +448,7 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
     data->ctxtVar = oldCtxtVar;
     if (params != NULL)
 	xsltFreeStackElemList(params);
+    tctxt->xpathCtxt->node = oldXPNode;
 
     if (data->error != 0)
         goto error;
@@ -468,20 +471,18 @@ exsltFuncFunctionFunction (xmlXPathParserContextPtr ctxt, int nargs) {
      * the generation of result nodes.
      */
     if (fake->children != NULL) {
-#ifdef LIBXML_DEBUG_ENABLED
-	xmlDebugDumpNode (stderr, fake, 1);
-#endif
 	xsltGenericError(xsltGenericErrorContext,
 			 "{%s}%s: cannot write to result tree while "
 			 "executing a function\n",
 			 ctxt->context->functionURI, ctxt->context->function);
-	xmlFreeNode(fake);
+        xmlXPathFreeObject(ret);
 	goto error;
     }
-    xmlFreeNode(fake);
     valuePush(ctxt, ret);
 
 error:
+    xmlFree(args);
+    xmlFreeNode(fake);
     tctxt->depth--;
 }
 
@@ -669,7 +670,7 @@ exsltFuncResultComp (xsltStylesheetPtr style, xmlNodePtr inst,
      */
     sel = xmlGetNsProp(inst, (const xmlChar *) "select", NULL);
     if (sel != NULL) {
-	ret->select = xmlXPathCompile (sel);
+	ret->select = xsltXPathCompileFlags(style, sel, 0);
 	xmlFree(sel);
     }
     /*
@@ -776,7 +777,7 @@ exsltFuncResultElem (xsltTransformContextPtr ctxt,
 	}
         /* Mark as function result. */
         xsltRegisterLocalRVT(ctxt, container);
-        container->psvi = XSLT_RVT_FUNC_RESULT;
+        container->compression = XSLT_RVT_FUNC_RESULT;
 
 	oldInsert = ctxt->insert;
 	ctxt->insert = (xmlNodePtr) container;
